from llvmlite import ir

from lleaves.compiler.utils import MissingType

BOOL = ir.IntType(bits=1)
DOUBLE = ir.DoubleType()
FLOAT = ir.FloatType()
INT_CAT = ir.IntType(bits=32)
INT = ir.IntType(bits=32)
ZERO_V = ir.Constant(BOOL, 0)
FLOAT_POINTER = ir.PointerType(FLOAT)
DOUBLE_PTR = ir.PointerType(DOUBLE)


def iconst(value):
    return ir.Constant(INT, value)


def fconst(value):
    return ir.Constant(FLOAT, value)


def dconst(value):
    return ir.Constant(DOUBLE, value)


def scalar_func(cat_bitmap):
    return ir.FunctionType(
        DOUBLE, (INT_CAT if is_cat else DOUBLE for is_cat in cat_bitmap)
    )


def ir_from_ast(forest):
    module = ir.Module(name="forest")

    # entry function called from Python
    root_func = ir.Function(
        module,
        ir.FunctionType(ir.VoidType(), (DOUBLE_PTR, DOUBLE_PTR, INT, INT)),
        name="forest_root",
    )

    tree_funcs = []
    for tree in forest.trees:
        # Declare the function for this tree
        tree_func = ir.Function(
            module, scalar_func(tree.categorical_bitmap), name=str(tree)
        )
        # add IR
        gen_tree(tree, tree_func)
        tree_funcs.append(tree_func)
    populate_forest_func(forest, root_func, tree_funcs)

    return module


def gen_tree(tree, tree_func):
    """generate code for tree given the function, recursing into nodes"""
    node_block = tree_func.append_basic_block(name=str(tree.root_node))
    gen_node(tree_func, node_block, tree.root_node)


def gen_node(func, node_block, node):
    """generate code for node, recursing into children"""
    if node.is_leaf():
        gen_leaf_node(node_block, node)
    else:
        gen_decision_node(func, node_block, node)


def gen_leaf_node(node_block, leaf):
    """populate block with leaf's return value"""
    builder = ir.IRBuilder(node_block)
    builder.ret(dconst(leaf.value))


def gen_decision_node(func, node_block, node):
    """generate code for decision node, recursing into children"""
    builder = ir.IRBuilder(node_block)

    # optimization for node where both children are leaves (switch instead of cbranch)
    is_fused_double_leaf_node = node.left.is_leaf() and node.right.is_leaf()
    if is_fused_double_leaf_node:
        left_block = None
        right_block = None
        # categorical nodes have a fastpath which can branch-right early
        # so they still need a right block
        if node.decision_type.is_categorical:
            right_block = func.append_basic_block(name=str(node.right))
    else:
        left_block = func.append_basic_block(name=str(node.left))
        right_block = func.append_basic_block(name=str(node.right))

    # If missingType != MNaN, LightGBM treats NaNs values as if they were 0.0.
    # So for MZero, NaNs get treated like missing values.
    # But for MNone, NaNs get treated as the literal value 0.0.
    # default_left decides where to go when a missing value is encountered

    if node.decision_type.is_categorical:
        bitset_comp_block = builder.append_basic_block("cat_bitset_comp")
        bitset_builder = ir.IRBuilder(bitset_comp_block)
        comp = populate_categorical_node_block(
            func, builder, bitset_builder, node, bitset_comp_block, right_block
        )
        builder = bitset_builder
    else:
        comp = populate_numerical_node_block(func, builder, node)

    if is_fused_double_leaf_node:
        ret = builder.select(comp, dconst(node.left.value), dconst(node.right.value))
        builder.ret(ret)
    else:
        builder.cbranch(comp, left_block, right_block)

    if left_block:
        gen_node(func, left_block, node.left)
    if right_block:
        gen_node(func, right_block, node.right)


def populate_forest_func(forest, root_func, tree_funcs):
    """Populate root function IR for forest"""
    data_arr, out_arr, start_index, end_index = root_func.args

    # -- SETUP BLOCK
    setup_block = root_func.append_basic_block("setup")
    builder = ir.IRBuilder(setup_block)
    loop_iter = builder.alloca(INT, 1, "loop-idx")
    builder.store(start_index, loop_iter)
    condition_block = root_func.append_basic_block("loop-condition")
    builder.branch(condition_block)
    # -- END SETUP BLOCK

    # -- CONDITION BLOCK
    builder = ir.IRBuilder(condition_block)
    comp = builder.icmp_signed("<", builder.load(loop_iter), end_index)
    core_block = root_func.append_basic_block("loop-core")
    term_block = root_func.append_basic_block("term")
    builder.cbranch(comp, core_block, term_block)
    # -- END CONDITION BLOCK

    # -- CORE LOOP BLOCK
    builder = ir.IRBuilder(core_block)
    # build args arr, convert categoricals vars from float to int
    args = []
    loop_iter_reg = builder.load(loop_iter)

    n_args = ir.Constant(INT, forest.n_args)
    iter_mul_nargs = builder.mul(loop_iter_reg, n_args)
    idx = (builder.add(iter_mul_nargs, iconst(i)) for i in range(forest.n_args))
    raw_ptrs = [builder.gep(root_func.args[0], (c,)) for c in idx]
    for is_cat, ptr in zip(forest.categorical_bitmap, raw_ptrs):
        el = builder.load(ptr)
        if is_cat:
            args.append(builder.fptosi(el, INT_CAT))
        else:
            args.append(el)
    # iterate over each tree, sum up results
    res = builder.call(tree_funcs[0], args)
    for func in tree_funcs[1:]:
        # could be inlined, but optimizer does for us
        tree_res = builder.call(func, args)
        res = builder.fadd(tree_res, res)
    ptr = builder.gep(out_arr, (loop_iter_reg,))
    builder.store(res, ptr)
    tmpp1 = builder.add(loop_iter_reg, iconst(1))
    builder.store(tmpp1, loop_iter)
    builder.branch(condition_block)
    # -- END CORE LOOP BLOCK

    # -- TERMINAL BLOCK
    ir.IRBuilder(term_block).ret_void()
    # -- END TERMINAL BLOCK


def populate_categorical_node_block(
    func, builder, bitset_comp_builder, node, bitset_comp_block, right_block
):
    """Populate block with IR for categorical node"""
    val = func.args[node.split_feature]

    # For categoricals, processing NaNs happens through casting them via fptosi in the Forest root
    # NaNs become negative max_val, which never exists in the Bitset, so they always go right

    # Find in bitset
    # First, check value > max categorical
    comp = builder.icmp_unsigned(
        "<",
        val,
        iconst(32 * len(node.cat_threshold)),
    )
    builder.cbranch(comp, bitset_comp_block, right_block)

    idx = bitset_comp_builder.udiv(val, iconst(32))
    bit_vecs = ir.Constant(
        ir.VectorType(INT, len(node.cat_threshold)),
        [ir.Constant(INT, i) for i in node.cat_threshold],
    )
    shift = bitset_comp_builder.urem(val, iconst(32))
    # pick relevant bitvector
    bit_vec = bitset_comp_builder.extract_element(bit_vecs, idx)
    # check bitvector contains
    bit_entry = bitset_comp_builder.lshr(bit_vec, shift)
    comp = bitset_comp_builder.trunc(bit_entry, BOOL)
    return comp


def populate_numerical_node_block(func, builder, node):
    """populate block with IR for numerical node"""
    val = func.args[node.split_feature]

    thresh = ir.Constant(DOUBLE, node.threshold)
    missing_t = node.decision_type.missing_type

    # for MNone handle NaNs by adjusting default_left to make sure NaNs go where 0.0 would have gone.
    # for MZero we handle NaNs in the IR
    if node.decision_type.missing_type == MissingType.MNone:
        default_left = 0.0 <= node.threshold
    else:
        default_left = node.decision_type.is_default_left

    # MissingType.MZero: Treat 0s (and NaNs) as missing values
    if default_left:
        if missing_t != MissingType.MZero or (
            missing_t == MissingType.MZero and 0.0 <= node.threshold
        ):
            # unordered cmp: we'll get True (and go left) if any arg is qNaN
            comp = builder.fcmp_unordered("<=", val, thresh)
        else:
            is_missing = builder.fcmp_unordered("==", val, fconst(0.0))
            less_eq = builder.fcmp_unordered("<=", val, thresh)
            comp = builder.or_(is_missing, less_eq)
    else:
        if missing_t != MissingType.MZero or (
            missing_t == MissingType.MZero and node.threshold < 0.0
        ):
            # ordered cmp: we'll get False (and go right) if any arg is qNaN
            comp = builder.fcmp_ordered("<=", val, thresh)
        else:
            is_missing = builder.fcmp_unordered("==", val, fconst(0.0))
            greater = builder.fcmp_ordered(">", val, thresh)
            comp = builder.not_(builder.or_(is_missing, greater))
    return comp



import json
import os


def scan_pandas_categorical(file_path):
    pandas_key = "pandas_categorical:"
    offset = -len(pandas_key)
    max_offset = -os.path.getsize(file_path)
    # seek backwards from end of file until we have to lines
    # the (pen)ultimate line should be pandas_categorical:XXX
    with open(file_path, "rb") as f:
        while True:
            if offset < max_offset:
                offset = max_offset
            f.seek(offset, os.SEEK_END)
            lines = f.readlines()
            if len(lines) >= 2:
                break
            offset *= 2
    last_line = lines[-1].decode().strip()
    if not last_line.startswith(pandas_key):
        last_line = lines[-2].decode().strip()
    if last_line.startswith(pandas_key):
        return json.loads(last_line[len(pandas_key) :])
    raise ValueError("Ill formatted model file!")


def scan_model_file(file_path, general_info_only=False):
    res = {"trees": []}

    with open(file_path, "r") as f:
        # List of blocks we expect:
        # 1* General Information
        # N* Tree, one block for each tree
        # 1* 'end of trees'
        # 1* Feature importances
        # 1* Parameters
        # 1* 'end of parameters'
        # 1* 'pandas_categorical:XXXXX'
        lines = _get_next_block_of_lines(f)
        assert lines[0] == "tree" and lines[1].startswith(
            "version="
        ), f"{file_path} is not a LightGBM model file"
        res["general_info"] = _struct_from_block(lines, INPUT_SCAN_KEYS)
        if general_info_only:
            return res

        lines = _get_next_block_of_lines(f)
        while lines:
            if lines[0].startswith("Tree="):
                res["trees"].append(_scan_tree(lines))
            else:
                assert lines[0] == "end of trees"
                break
            lines = _get_next_block_of_lines(f)
    res["pandas_categorical"] = scan_pandas_categorical(file_path)
    return res


def _scan_tree(lines):
    struct = _struct_from_block(lines, TREE_SCAN_KEYS)
    return struct


def _get_next_block_of_lines(file):
    # the only function where we advance file_offset
    result = []
    line = file.readline()
    while line == "\n":
        line = file.readline()
    while line != "\n" and line != "":
        result.append(line.strip())
        line = file.readline()
    return result


def cat_args_bitmap(arr):
    # Feature infos for floats look like [x.xxxx:y.yyyy]
    # for categoricals like X:Y:Z:
    return [not val.startswith("[") for val in arr]


class ScannedValue:
    def __init__(self, type: type, is_list=False, null_ok=False):
        self.type = type
        self.is_list = is_list
        self.null_ok = null_ok


INPUT_SCAN_KEYS = {
    "max_feature_idx": ScannedValue(int),
    "version": ScannedValue(str),
    "feature_infos": ScannedValue(str, True),
    "objective": ScannedValue(str, True),
}
TREE_SCAN_KEYS = {
    "Tree": ScannedValue(int),
    "num_leaves": ScannedValue(int),
    "num_cat": ScannedValue(int),
    "split_feature": ScannedValue(int, True),
    "threshold": ScannedValue(float, True),
    "decision_type": ScannedValue(int, True),
    "left_child": ScannedValue(int, True),
    "right_child": ScannedValue(int, True),
    "leaf_value": ScannedValue(float, True),
    "cat_threshold": ScannedValue(int, True, True),
    "cat_boundaries": ScannedValue(int, True, True),
}


def _struct_from_block(lines: list, keys_to_scan: dict):
    """
    Scans a block (= list of lines), produces a key: value struct
    @param lines: list of lines in the block
    @param keys_to_scan: dict with 'key': 'type of value' of keys to scan for
    """
    struct = {}
    for line in lines:
        # initial line in file
        if line == "tree":
            continue

        key, value = line.split("=")
        if key in keys_to_scan.keys():
            value_type = keys_to_scan[key]
            if value_type.is_list:
                if value:
                    parsed_value = [value_type.type(x) for x in value.split(" ")]
                else:
                    parsed_value = []
            else:
                parsed_value = value_type.type(value)
            struct[key] = parsed_value

    missing_keys = keys_to_scan.keys() - struct.keys()
    for key in missing_keys:
        value = keys_to_scan[key]
        assert value.null_ok, f"Non-nullable key {key} wasn't found"
        struct[key] = None
    return struct



import concurrent.futures
import os
from ctypes import CFUNCTYPE, POINTER, c_double, c_int
from pathlib import Path

import llvmlite.binding as llvm
import numpy as np

from lleaves import compiler
from lleaves.compiler.ast import scanner
from lleaves.compiler.objective_funcs import get_objective_func

try:
    from pandas import DataFrame as pd_DataFrame
except ImportError:

    class pd_DataFrame:
        """Dummy class for pandas.DataFrame."""

        pass


class Model:
    """
    The base class of lleaves.
    """

    # machine-targeted compiler & exec engine.
    _execution_engine = None

    # LLVM IR Module
    _IR_module: llvm.ModuleRef = None

    # prediction function, drops GIL on entry
    _c_entry_func = None

    def __init__(self, model_file=None):
        """
        Initialize the uncompiled model.

        :param model_file: Path to the model.txt.
        """
        self.model_file = model_file
        self.is_compiled = False

        scanned_model = scanner.scan_model_file(model_file, general_info_only=True)
        self._general_info = scanned_model["general_info"]
        self._pandas_categorical = scanner.scan_pandas_categorical(model_file)

        # objective function is implemented as an np.ufunc.
        self.objective_transf = get_objective_func(self._general_info["objective"])

    def num_feature(self):
        """
        Returns the number of features used by this model.
        """
        return self._general_info["max_feature_idx"] + 1

    def _get_execution_engine(self):
        """
        Create an empty ExecutionEngine suitable for JIT code generation on
        the host CPU. The engine is reusable for an arbitrary number of
        modules.
        """
        if self._execution_engine:
            return self._execution_engine

        llvm.initialize()
        llvm.initialize_native_target()
        llvm.initialize_native_asmprinter()

        # Create a target machine representing the host
        target = llvm.Target.from_default_triple()
        target_machine = target.create_target_machine()
        # And an execution engine with an empty backing module
        backing_mod = llvm.parse_assembly("")
        self._execution_engine = llvm.create_mcjit_compiler(backing_mod, target_machine)
        return self._execution_engine

    def _get_llvm_module(self):
        if self._IR_module:
            return self._IR_module
        self._IR_module = compiler.compile_to_module(self.model_file)
        return self._IR_module

    def compile(self, cache=None):
        """
        Generate the LLVM IR for this model and compile it to ASM.

        This function can be called multiple times, but will only compile once.

        :param cache: Path to a cache file. If this path doesn't exist, binary will be dumped at path after compilation.
                      If path exists, binary will be loaded and compilation skipped.
                      No effort is made to check staleness / consistency.
                      The precise workings of the cache parameter will be subject to future changes.
        """
        if self.is_compiled:
            return

        llvm.initialize()
        llvm.initialize_native_target()
        llvm.initialize_native_asmprinter()

        # Create a target machine representing the host
        target = llvm.Target.from_default_triple()
        target_machine = target.create_target_machine()

        if cache is None or not Path(cache).exists():
            # Compile to LLVM IR
            module = self._get_llvm_module()
        else:
            # when loading binary from cache we use a dummy empty module
            module = llvm.parse_assembly("")

        # Create execution engine for our module
        self._execution_engine = llvm.create_mcjit_compiler(module, target_machine)

        # when caching we dump the executable once the module finished compiling
        def save_to_cache(module, buffer):
            if cache and not Path(cache).exists():
                with open(cache, "wb") as file:
                    file.write(buffer)

        # when caching load the executable if it exists
        def load_from_cache(module):
            if cache and Path(cache).exists():
                return Path(cache).read_bytes()

        self._execution_engine.set_object_cache(
            notify_func=save_to_cache, getbuffer_func=load_from_cache
        )

        # compile IR to ASM
        self._execution_engine.finalize_object()
        self._execution_engine.run_static_constructors()

        # construct entry func
        addr = self._execution_engine.get_function_address("forest_root")
        # CFUNCTYPE params: void return, pointer to data, pointer to results arr, start_idx, end_idx
        # Drops GIL during call, re-aquires it after
        self._c_entry_func = CFUNCTYPE(
            None, POINTER(c_double), POINTER(c_double), c_int, c_int
        )(addr)
        self.is_compiled = True

    def predict(self, data, n_jobs=os.cpu_count()):
        """
        Return predictions for the given data.

        The model needs to be compiled before prediction.

        :param data: Pandas df, numpy 2D array or Python list. For fastest speed pass 2D float64 numpy arrays only.
        :param n_jobs: Number of threads to use for prediction. Defaults to number of CPUs. For single-row prediction
            this should be set to 1.
        :return: 1D numpy array (dtype float64)
        """
        if not self.is_compiled:
            raise RuntimeError(
                "Model needs to be compiled before prediction. Run model.compile()."
            )

        if isinstance(data, pd_DataFrame):
            data = self._data_from_pandas(data)
        data, n_preds = self._to_1d_ndarray(data)
        ptr_data = data.ctypes.data_as(POINTER(c_double))

        preds = np.zeros(n_preds, dtype=np.float64)
        ptr_preds = preds.ctypes.data_as(POINTER(c_double))
        if n_jobs > 1:
            batchsize = n_preds // n_jobs + (n_preds % n_jobs > 0)

            def f(start_idx):
                self._c_entry_func(
                    ptr_data, ptr_preds, start_idx, min(start_idx + batchsize, n_preds)
                )

            with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
                for i in range(0, n_preds, batchsize):
                    executor.submit(f, i)
        else:
            self._c_entry_func(ptr_data, ptr_preds, 0, n_preds)
        return self.objective_transf(preds)

    def _data_from_pandas(self, data):
        if len(data.shape) != 2 or data.shape[0] < 1:
            raise ValueError("Input data must be 2D and non-empty.")
        cat_cols = list(data.select_dtypes(include=["category"]).columns)
        if len(cat_cols) != len(self._pandas_categorical):
            print(cat_cols, self._pandas_categorical)
            raise ValueError(
                "The categorical features passed don't match the train dataset."
            )
        for col, category in zip(cat_cols, self._pandas_categorical):
            # we use set_categories to get the same (category -> code) mapping that we used during train
            if list(data[col].cat.categories) != list(category):
                data[col] = data[col].cat.set_categories(category)
        if len(cat_cols):  # cat_cols is list
            data = data.copy()
            # apply (category -> code) mapping. Categories become floats
            data[cat_cols] = (
                data[cat_cols].apply(lambda x: x.cat.codes).replace({-1: np.nan})
            )
        data = data.values
        if data.dtype != np.float32 and data.dtype != np.float64:
            data = data.astype(np.float64)
        return data

    def _to_1d_ndarray(self, data):
        if isinstance(data, list):
            try:
                data = np.array(data, dtype=np.float64)
            except BaseException:
                raise ValueError("Cannot convert data list to appropriate np array")

        if not isinstance(data, np.ndarray):
            raise ValueError(f"Expecting list or numpy.ndarray, got {type(data)}")
        if len(data.shape) != 2:
            raise ValueError(
                f"Data must be 2 dimensional, is {len(data.shape)} dimensional"
            )
        n_preds = data.shape[0]
        if data.dtype == np.float64:
            # flatten the array to 1D
            data = np.array(data.reshape(data.size), dtype=np.float64, copy=False)
        else:
            data = np.array(data.reshape(data.size), dtype=np.float64)
        return data, n_preds



import numpy as np

from lleaves.compiler.utils import ISSUE_ERROR_MSG


def _get_sigmoid(alpha):
    if alpha <= 0:
        raise ValueError(f"Sigmoid parameter needs to be >0, is {alpha}")

    # Numerically stable implementation of 1 / (1 + exp(-αlpha*x))
    return lambda x: 0.5 * (1 + np.tanh(0.5 * x * alpha))


def get_objective_func(obj):
    """
    Given the objective=obj entry in the model.txt, return callable
    :return: callable implementing the correct result transformation as np.ufunc
    """
    if obj[0] == "binary":
        func, alpha = obj[1].split(":")
        f = _get_sigmoid(float(alpha))
    elif obj[0] in ("xentropy", "cross_entropy"):
        f = _get_sigmoid(1.0)
    elif obj[0] in ("xentlambda", "cross_entropy_lambda"):

        def f(x):
            return np.log1p(np.exp(x))

    elif obj[0] in ("poisson", "gamma", "tweedie"):
        f = np.exp
    elif obj[0] in ("regression", "regression_l1", "huber", "fair", "quantile", "mape"):
        if "sqrt" in obj[1:]:

            def f(x):
                return np.copysign(np.square(x), x)

        else:

            def f(x):
                return x

    elif obj[0] in ("lambdarank", "rank_xendcg", "custom"):

        def f(x):
            return x

    else:
        raise ValueError(f"Objective '{obj[0]}' not yet implemented. {ISSUE_ERROR_MSG}")
    return f



import llvmlite.binding as llvm

from lleaves.compiler.ast import parse_to_ast
from lleaves.compiler.codegen import ir_from_ast


def compile_to_module(file_path):
    forest = parse_to_ast(file_path)
    ir = ir_from_ast(forest)

    module = llvm.parse_assembly(str(ir))
    module.verify()

    # Create optimizer
    pmb = llvm.PassManagerBuilder()
    pmb.opt_level = 3
    pmb.inlining_threshold = 30
    pm_module = llvm.ModulePassManager()
    # Add optimization passes to module-level optimizer
    pmb.populate(pm_module)

    # single pass only, compiler optimizations don't bring much speedup and take time
    pm_module.run(module)
    return module



import pickle
import time
from pathlib import Path
from statistics import mean, pstdev

import lightgbm
import matplotlib.pyplot as plt
import numpy as np
import onnxmltools
import onnxruntime as rt
import pandas as pd
import seaborn
import treelite
import treelite_runtime
from onnxconverter_common import FloatTensorType

from benchmarks.train_NYC_model import feature_enginering
from lleaves import Model


class BenchmarkModel:
    model = None
    name = None

    def __init__(self, lgbm_model_file):
        self.model_file = lgbm_model_file

    def setup(self, data, n_threads):
        start = time.perf_counter()
        self._setup(data, n_threads)
        print(f"{self.name} setup: {round(time.perf_counter() - start, 2)}")

    def _setup(self, data, n_threads):
        raise NotImplementedError()

    def predict(self, data, index, batchsize, n_threads):
        return self.model.predict(data[index : index + batchsize], n_jobs=n_threads)

    def __str__(self):
        return self.name


class LGBMModel(BenchmarkModel):
    name = "LightGBM Booster"

    def _setup(self, data, n_threads):
        self.model = lightgbm.Booster(model_file=self.model_file)


class LLVMModel(BenchmarkModel):
    name = "lleaves"

    def _setup(self, data, n_threads):
        self.model = Model(model_file=self.model_file)
        self.model.compile()


class LLVMModelSingle(BenchmarkModel):
    name = "lleaves (single-threaded)"

    def _setup(self, data, n_threads):
        self.model = Model(model_file=self.model_file)
        self.model.compile()

    def predict(self, data, index, batchsize, n_threads):
        return self.model.predict(data[index : index + batchsize], n_jobs=1)


class TreeliteModel(BenchmarkModel):
    name = "Treelite"

    def _setup(self, data, n_threads):
        treelite_model = treelite.Model.load(self.model_file, model_format="lightgbm")
        treelite_model.export_lib(
            toolchain="gcc",
            libpath="/tmp/treelite_model.so",
            params={"parallel_comp": 4},
            verbose=False,
        )
        self.model = treelite_runtime.Predictor(
            "/tmp/treelite_model.so",
            nthread=n_threads,
        )

    def predict(self, data, index, batchsize, n_threads):
        return self.model.predict(
            treelite_runtime.DMatrix(data[index : index + batchsize])
        )


class ONNXModel(BenchmarkModel):
    name = "ONNX Runtime"

    def _setup(self, data, n_threads):
        lgbm_model = lightgbm.Booster(model_file=self.model_file)
        onnx_model = onnxmltools.convert_lightgbm(
            lgbm_model,
            initial_types=[
                (
                    "float_input",
                    FloatTensorType([None, lgbm_model.num_feature()]),
                )
            ],
            target_opset=8,
        )
        onnxmltools.utils.save_model(onnx_model, "/tmp/model.onnx")
        options = rt.SessionOptions()
        options.inter_op_num_threads = n_threads
        options.intra_op_num_threads = n_threads
        self.model = rt.InferenceSession("/tmp/model.onnx", sess_options=options)
        self.input_name = self.model.get_inputs()[0].name
        self.label_name = self.model.get_outputs()[0].name

    def predict(self, data, index, batchsize, n_threads):
        return self.model.run(
            [self.label_name], {self.input_name: data[index : index + batchsize]}
        )


def get_color(libname):
    libkey = libname.lower()
    d = {
        "lleaves": "red",
        "lightgbm": "sandybrown",
        "treelite": "mediumseagreen",
        "onnx": "cornflowerblue",
    }
    for key, value in d.items():
        if key in libkey:
            return value


def save_plots(results_full, n_threads, model_files, batchsizes):
    for n_thread in n_threads:
        fig, axs = plt.subplots(ncols=2, nrows=1)
        fig.set_size_inches(16, 9)
        keys = sorted(results_full.keys())
        for count, model_file in enumerate(model_files):
            model_name = model_file.split("/")[-2]
            for key in keys:
                if key.startswith(f"{model_name}_{n_thread}"):
                    if "(single-threaded)" in key:
                        continue
                    seaborn.lineplot(
                        x="batchsize",
                        y="time (μs)",
                        ci="sd",
                        data=results_full[key],
                        ax=axs[count],
                        color=get_color(key),
                        label=key.split("_")[3] if count == 0 else None,
                    )
            axs[count].set(
                xscale="log",
                yscale="log",
                title=model_name,
                xticks=batchsizes,
                xticklabels=batchsizes,
                xlim=(1, None) if n_thread == 1 else None,
            )
        plt.savefig(f"{n_thread}.png", bbox_inches="tight")


NYC_used_columns = [
    "fare_amount",
    "pickup_latitude",
    "pickup_longitude",
    "dropoff_latitude",
    "dropoff_longitude",
    "tpep_pickup_datetime",
    "passenger_count",
]


def load_results():
    datafile = Path("data/results.pkl")
    res = {}
    if datafile.exists():
        with open(datafile, "rb") as file:
            res = pickle.load(file)
    return res


def save_results(results):
    datafile = Path("data/results.pkl")
    with open(datafile, "wb") as file:
        pickle.dump(results, file)


def run_benchmark(
    model_files, np_data, model_classes, threadcount, batchsizes, n_samples=1000
):
    results_full = load_results()
    for model_file, data in zip(model_files, np_data):
        model_name = model_file.split("/")[-2]
        print(model_file, f"\n---- {str.upper(model_name)} --- \n")
        for n_threads in threadcount:
            for model_class in model_classes:
                model = model_class(model_file)
                results = {"time (μs)": [], "batchsize": []}
                results_full[f"{model_name}_{n_threads}_{model}"] = results
                model.setup(data, n_threads)
                for batchsize in batchsizes:
                    times = []
                    for _ in range(n_samples):
                        start = time.perf_counter_ns()
                        model.predict(data, 0, batchsize, n_threads)
                        # calc per-batch times, in μs
                        times.append((time.perf_counter_ns() - start) / 1000)
                    results["time (μs)"] += times
                    results["batchsize"] += len(times) * [batchsize]
                    print(
                        f"{model} (Batchsize {batchsize}, nthread {n_threads}): {round(mean(times), 2)}μs ± {round(pstdev(times), 2)}μs"
                    )
    save_results(results_full)
    save_plots(results_full, threadcount, model_files, batchsizes)


if __name__ == "__main__":
    df = pd.read_parquet(
        "data/yellow_tripdata_2016-01.parquet", columns=NYC_used_columns
    )
    NYC_X = feature_enginering().fit_transform(df).astype(np.float32)

    df = pd.read_csv("data/airline_data_factorized.csv")
    airline_X = df.to_numpy(np.float32)

    model_file_NYC = "../tests/models/NYC_taxi/model.txt"
    model_file_airline = "../tests/models/airline/model.txt"

    run_benchmark(
        model_files=[model_file_NYC, model_file_airline],
        np_data=[None, None],
        model_classes=[
            LLVMModel,
            LLVMModelSingle,
            LGBMModel,
            TreeliteModel,
            ONNXModel,
        ],
        threadcount=[4],
        batchsizes=[
            1000,
            3000,
            5000,
            10000,
            100000,
            300000,
            1000000,
        ],
        n_samples=100,
    )

    run_benchmark(
        model_files=[model_file_NYC, model_file_airline],
        np_data=[NYC_X, airline_X],
        model_classes=[
            LLVMModel,
            LGBMModel,
            TreeliteModel,
            ONNXModel,
        ],
        threadcount=[1],
        batchsizes=[1, 2, 3, 5, 7, 10, 20, 30, 50, 70, 100, 200, 300],
        n_samples=20000,
    )



from lleaves.compiler.ast.nodes import DecisionNode, Forest, LeafNode, Tree
from lleaves.compiler.ast.scanner import cat_args_bitmap, scan_model_file
from lleaves.compiler.utils import DecisionType


def _parse_tree_to_ast(tree_struct, cat_bitmap):
    n_nodes = len(tree_struct["decision_type"])
    leaves = [
        LeafNode(idx, value) for idx, value in enumerate(tree_struct["leaf_value"])
    ]

    # Create the nodes using all non-specific data
    # categorical nodes are finalized later
    nodes = [
        DecisionNode(
            idx, split_feature, threshold, decision_type_id, left_idx, right_idx
        )
        for idx, (
            split_feature,
            threshold,
            decision_type_id,
            left_idx,
            right_idx,
        ) in enumerate(
            zip(
                tree_struct["split_feature"],
                tree_struct["threshold"],
                tree_struct["decision_type"],
                tree_struct["left_child"],
                tree_struct["right_child"],
            )
        )
    ]
    assert len(nodes) == n_nodes

    categorical_nodes = [
        idx
        for idx, decision_type_id in enumerate(tree_struct["decision_type"])
        if DecisionType(decision_type_id).is_categorical
    ]

    for idx in categorical_nodes:
        node = nodes[idx]
        thresh = int(node.threshold)
        # pass just the relevant vector entries
        start = tree_struct["cat_boundaries"][thresh]
        end = tree_struct["cat_boundaries"][thresh + 1]
        node.finalize_categorical(
            cat_threshold=tree_struct["cat_threshold"][start:end],
        )

    for node in nodes:
        # in the model_file.txt, the outgoing left + right nodes are specified
        # via their index in the list. negative numbers are leaves, positive numbers
        # are other nodes
        children = [
            leaves[abs(idx) - 1] if idx < 0 else nodes[idx]
            for idx in (node.left_idx, node.right_idx)
        ]
        node.add_children(*children)

    for node in nodes:
        node.validate()

    if nodes:
        return Tree(tree_struct["Tree"], nodes[0], cat_bitmap)
    else:
        # special case for when tree is just single leaf
        assert len(leaves) == 1
        return Tree(tree_struct["Tree"], leaves[0], cat_bitmap)


def parse_to_ast(model_path):
    scanned_model = scan_model_file(model_path)
    n_args = scanned_model["general_info"]["max_feature_idx"] + 1
    cat_bitmap = cat_args_bitmap(scanned_model["general_info"]["feature_infos"])
    assert n_args == len(cat_bitmap), "Ill formed model file"

    trees = [
        _parse_tree_to_ast(tree_struct, cat_bitmap)
        for tree_struct in scanned_model["trees"]
    ]
    return Forest(trees, cat_bitmap)



from pathlib import Path

from lleaves.compiler.ast.scanner import scan_model_file, scan_pandas_categorical


def test_parser():
    model_file = "tests/models/boston_housing/model.txt"
    result = scan_model_file(model_file)
    assert result["general_info"]["max_feature_idx"] == 12
    assert result["pandas_categorical"] is None
    assert len(result["trees"]) == 100

    tree_3 = result["trees"][3]
    assert tree_3["num_leaves"] == 18
    assert tree_3["left_child"] == [
        1,
        3,
        -2,
        -1,
        9,
        8,
        7,
        10,
        15,
        16,
        -5,
        -9,
        -8,
        14,
        -14,
        -6,
        -3,
    ]

    tree_95 = result["trees"][95]
    assert tree_95["Tree"] == 95
    assert tree_95["num_leaves"] == 21

    model_file = "tests/models/tiniest_single_tree/model.txt"
    result = scan_model_file(model_file)
    assert len(result["trees"]) == 1
    tree_0 = result["trees"][0]
    assert tree_0["num_leaves"] == 4
    assert result["pandas_categorical"] is None


def test_parsing_pandas(tmp_path):
    mod_model_file = tmp_path / "mod_model.txt"
    model_file = Path("tests/models/pure_categorical/model.txt")
    with open(model_file, "r") as file:
        lines = file.readlines()
    assert lines[-1].startswith("pandas_categorical")
    lines[
        -1
    ] = 'pandas_categorical:[["a", "b", "c"], ["b", "c", "d"], ["w", "x", "y", "z"]]'

    with open(mod_model_file, "x") as file:
        file.writelines(lines)

    pandas_categorical = scan_pandas_categorical(model_file)
    assert pandas_categorical is None
    pandas_categorical = scan_pandas_categorical(mod_model_file)
    assert pandas_categorical == [
        ["a", "b", "c"],
        ["b", "c", "d"],
        ["w", "x", "y", "z"],
    ]
